@@ -1,3 +1,5 @@
+/* -*- mode: C; indent-tabs-mode: t; tab-width: 8; c-basic-offset: 2; -*- */
+
 /*
  * This file is part of Seed, the GObject Introspection<->Javascript bindings.
  *
@@ -12,7 +14,7 @@
  * You should have received a copy of the GNU Lesser General Public License
  * along with Seed.  If not, see <http://www.gnu.org/licenses/>.
  *
- * Copyright (C) Robert Carr 2008 <carrr@rpi.edu>
+ * Copyright (C) Robert Carr 2009 <carrr@rpi.edu>
  */
 
 #include "seed-private.h"
@@ -20,8 +22,10 @@
 
 #include <sys/types.h>
 #include <sys/wait.h>
-#include <string.h>
 #include <stdarg.h>
+#include <string.h>
+
+JSObjectRef function_proto;
 
 JSObjectRef seed_obj_ref;
 
@@ -68,7 +72,8 @@ static const GDebugKey seed_debug_keys[] = {
  * to a fresh #SeedContext.
  *
  */
-void seed_prepare_global_context (JSContextRef ctx)
+void
+seed_prepare_global_context (JSContextRef ctx)
 {
   JSObjectRef global = JSContextGetGlobalObject (ctx);
 
@@ -199,7 +204,7 @@ seed_gobject_constructor_invoked (JSContextRef ctx,
 	{
 	  g_free (params);
 	  JSPropertyNameArrayRelease (jsprops);
-	  seed_next_gobject_wrapper =  NULL;
+	  seed_next_gobject_wrapper = NULL;
 	  return 0;
 	}
       params[ri].name = prop_name;
@@ -215,11 +220,10 @@ seed_gobject_constructor_invoked (JSContextRef ctx,
   gobject = g_object_newv (type, ri, params);
 
 
-  if (G_IS_INITIALLY_UNOWNED (gobject) &&
-      !g_object_is_floating(gobject))
-    g_object_ref(gobject);
-  else if (g_object_is_floating(gobject))
-    g_object_ref_sink(gobject);
+  if (G_IS_INITIALLY_UNOWNED (gobject) && !g_object_is_floating (gobject))
+    g_object_ref (gobject);
+  else if (g_object_is_floating (gobject))
+    g_object_ref_sink (gobject);
 
   if (!gobject)
     ret = (JSObjectRef) JSValueMakeNull (ctx);
@@ -291,15 +295,14 @@ seed_gobject_method_finalize (JSObjectRef method)
     g_base_info_unref (info);
 }
 
-typedef void (*InitMethodCallback) (gint *argc, gchar ***argv);
+typedef void (*InitMethodCallback) (gint * argc, gchar *** argv);
 
 static gboolean
 seed_gobject_init_build_argv (JSContextRef ctx,
 			      JSObjectRef array,
-			      SeedArgvPrivates *priv,
-			      JSValueRef *exception)
+			      SeedArgvPrivates * priv, JSValueRef * exception)
 {
-  guint i,length;
+  guint i, length;
   JSValueRef jsl;
 
   jsl = seed_object_get_property (ctx, array, "length");
@@ -307,13 +310,15 @@ seed_gobject_init_build_argv (JSContextRef ctx,
     return FALSE;
 
   length = seed_value_to_uint (ctx, jsl, exception);
-  priv->argv = g_new(gchar *, length);
+  priv->argv = g_new (gchar *, length);
   priv->argc = length;
 
   for (i = 0; i < length; i++)
     {
       priv->argv[i] = seed_value_to_string (ctx,
-					    JSObjectGetPropertyAtIndex (ctx, array, i,
+					    JSObjectGetPropertyAtIndex (ctx,
+									array,
+									i,
 									exception),
 					    exception);
     }
@@ -338,31 +343,33 @@ seed_gobject_init_method_invoked (JSContextRef ctx,
   if (argumentCount != 1 && argumentCount != 2)
     {
       seed_make_exception (ctx, exception,
-			   "ArgumentError", "init method expects 1 argument, got %zd",
+			   "ArgumentError",
+			   "init method expects 1 argument, got %zd",
 			   argumentCount);
       return JSValueMakeUndefined (ctx);
     }
 
-  if (argumentCount ==1)
+  if (argumentCount == 1)
     {
-      if (JSValueIsNull (ctx, arguments[0]) || !JSValueIsObject (ctx, arguments[0]))
+      if (JSValueIsNull (ctx, arguments[0])
+	  || !JSValueIsObject (ctx, arguments[0]))
 
 	{
 	  seed_make_exception (ctx, exception,
-			       "ArgumentError", "init method expects an array object as argument");
+			       "ArgumentError",
+			       "init method expects an array object as argument");
 	  return JSValueMakeUndefined (ctx);
 	}
-      if(JSValueIsObjectOfClass (ctx, arguments[0], seed_argv_class))
+      if (JSValueIsObjectOfClass (ctx, arguments[0], seed_argv_class))
 	{
-	  priv = JSObjectGetPrivate ((JSObjectRef)arguments[0]);
+	  priv = JSObjectGetPrivate ((JSObjectRef) arguments[0]);
 	}
       else
 	{
-	  priv = g_newa (SeedArgvPrivates,1);
+	  priv = g_newa (SeedArgvPrivates, 1);
 	  if (!seed_gobject_init_build_argv (ctx,
-					     (JSObjectRef)arguments[0],
-					     priv,
-					     exception))
+					     (JSObjectRef) arguments[0],
+					     priv, exception))
 	    {
 	      seed_make_exception (ctx, exception, "ArgumentError",
 				   "Init method expects an array as argument");
@@ -375,15 +382,17 @@ seed_gobject_init_method_invoked (JSContextRef ctx,
 
   info = JSObjectGetPrivate (function);
   typelib = g_base_info_get_typelib (info);
-  g_typelib_symbol (typelib, g_function_info_get_symbol ((GIFunctionInfo *)info), (gpointer *)&c);
+  g_typelib_symbol (typelib,
+		    g_function_info_get_symbol ((GIFunctionInfo *) info),
+		    (gpointer *) & c);
   // Backwards compatibility
   if (!priv)
     {
-      c(NULL, NULL);
+      c (NULL, NULL);
       return JSValueMakeUndefined (ctx);
     }
 
-  c(&priv->argc, &priv->argv);
+  c (&priv->argc, &priv->argv);
 
   if (allocated)
     g_free (priv->argv);
@@ -407,6 +416,7 @@ seed_gobject_method_invoked (JSContextRef ctx,
   GArgument *out_args;
   GArgument *out_values;
   guint n_args, n_in_args, n_out_args, i;
+  guint in_args_pos, out_args_pos;
   GIArgInfo *arg_info;
   GITypeInfo *type_info;
   GIDirection dir;
@@ -415,8 +425,7 @@ seed_gobject_method_invoked (JSContextRef ctx,
 
   info = JSObjectGetPrivate (function);
   // We just want to check if there IS an object, not actually throw an
-  // exception if we don't
-  // get it.
+  // exception if we don't get it.
   if (!this_object || !
       ((object = seed_value_to_object (ctx, this_object, 0)) ||
        (object = seed_pointer_get_pointer (ctx, this_object))))
@@ -434,7 +443,9 @@ seed_gobject_method_invoked (JSContextRef ctx,
 
   for (i = 0; (i < (n_args)); i++)
     {
-      SEED_NOTE (INVOCATION, "Converting arg: %d of function %s, exception is %p", i, g_base_info_get_name (info), exception);
+      SEED_NOTE (INVOCATION,
+		 "Converting arg: %d of function %s, exception is %p", i,
+		 g_base_info_get_name (info), exception);
       arg_info = g_callable_info_get_arg ((GICallableInfo *) info, i);
       dir = g_arg_info_get_direction (arg_info);
       type_info = g_arg_info_get_type (arg_info);
@@ -454,7 +465,8 @@ seed_gobject_method_invoked (JSContextRef ctx,
 				   "Unable to make argument %d for"
 				   " function: %s. \n",
 				   i + 1,
-				   g_base_info_get_name ((GIBaseInfo *) info));
+				   g_base_info_get_name ((GIBaseInfo *)
+							 info));
 
 	      g_base_info_unref ((GIBaseInfo *) type_info);
 	      g_base_info_unref ((GIBaseInfo *) arg_info);
@@ -543,6 +555,8 @@ seed_gobject_method_invoked (JSContextRef ctx,
 
       return JSValueMakeNull (ctx);
     }
+
+  in_args_pos = out_args_pos = 0;
   for (i = 0; (i < n_args); i++)
     {
       JSValueRef jsout_val;
@@ -554,14 +568,17 @@ seed_gobject_method_invoked (JSContextRef ctx,
 	{
 	  seed_gi_release_in_arg (g_arg_info_get_ownership_transfer
 				  (arg_info), type_info,
-				  &in_args[i + (instance_method ? 1 : 0)]);
+				  &in_args[in_args_pos +
+					   (instance_method ? 1 : 0)]);
+	  in_args_pos++;
 
 	  g_base_info_unref ((GIBaseInfo *) type_info);
 	  g_base_info_unref ((GIBaseInfo *) arg_info);
 	  continue;
 	}
-      jsout_val = seed_gi_argument_make_js (ctx, &out_values[i],
+      jsout_val = seed_gi_argument_make_js (ctx, &out_args[out_args_pos],
 					    type_info, exception);
+      out_args_pos++;
       if (!JSValueIsNull (ctx, arguments[i]) &&
 	  JSValueIsObject (ctx, arguments[i]))
 	{
@@ -615,8 +632,10 @@ seed_gobject_define_property_from_function_info (JSContextRef ctx,
   method_ref = JSObjectMake (ctx, gobject_method_class,
 			     g_base_info_ref ((GIBaseInfo *) info));
 
+  JSObjectSetPrototype (ctx, method_ref, function_proto);
+
   name = g_base_info_get_name ((GIBaseInfo *) info);
-  if (!strcmp (name, "new"))
+  if (!g_strcmp0 (name, "new"))
     name = "c_new";
   seed_object_set_property (ctx, object, name, method_ref);
   seed_object_set_property (ctx, method_ref, "info",
@@ -633,14 +652,13 @@ seed_gobject_add_methods_for_interfaces (JSContextRef ctx,
 					 JSObjectRef object)
 {
   GIInterfaceInfo *interface;
-  gint n_interfaces, i;
+  GIFunctionInfo *function;
+  gint n_interfaces, i, n_functions, k;
 
   n_interfaces = g_object_info_get_n_interfaces (oinfo);
 
   for (i = 0; i < n_interfaces; i++)
     {
-      GIFunctionInfo *function;
-      gint n_functions, k;
       interface = g_object_info_get_interface (oinfo, i);
 
       n_functions = g_interface_info_get_n_methods (interface);
@@ -729,8 +747,8 @@ seed_gobject_get_class_for_gtype (JSContextRef ctx, GType type)
       GType *interfaces;
       GIFunctionInfo *function;
       GIBaseInfo *interface;
-      gint n_functions, k, i;
-      guint n;
+      gint n_functions, k;
+      guint i, n;
 
       interfaces = g_type_interfaces (type, &n);
       for (i = 0; i < n; i++)
@@ -804,7 +822,8 @@ seed_gobject_get_property (JSContextRef context,
   char *cproperty_name;
   gint length;
   JSValueRef ret;
-  gint i, len;
+  guint i;
+  gsize len;
 
   b = seed_value_to_object (context, (JSValueRef) object, exception);
   if (!b)
@@ -819,9 +838,8 @@ seed_gobject_get_property (JSContextRef context,
 
   if (!spec)
     {
-
-      len = strlen (cproperty_name);
-      for (i = 0; i < len - 1; i++)
+      len = strlen (cproperty_name) - 1;
+      for (i = 0; i < len; i++)
 	{
 	  if (cproperty_name[i] == '_')
 	    cproperty_name[i] = '-';
@@ -832,13 +850,13 @@ seed_gobject_get_property (JSContextRef context,
 	goto found;
       else
 	{
-	  GIFieldInfo *field = 0;
+	  GIFieldInfo *field = NULL;
 	  GIBaseInfo *info = (GIBaseInfo *)
 	    g_irepository_find_by_gtype (0, G_OBJECT_TYPE (b));
 	  gint n;
 	  const gchar *name;
 
-	  for (i = 0; i < len - 1; i++)
+	  for (i = 0; i < len; i++)
 	    {
 	      if (cproperty_name[i] == '-')
 		cproperty_name[i] = '_';
@@ -846,7 +864,7 @@ seed_gobject_get_property (JSContextRef context,
 
 	  if (!info)
 	    {
-	      return 0;
+	      return NULL;
 	    }
 
 	  n = g_object_info_get_n_fields ((GIObjectInfo *) info);
@@ -855,7 +873,7 @@ seed_gobject_get_property (JSContextRef context,
 	      field = g_object_info_get_field ((GIObjectInfo *) info, i);
 	      name = g_base_info_get_name ((GIBaseInfo *) field);
 
-	      if (!strcmp (name, cproperty_name))
+	      if (!g_strcmp0 (name, cproperty_name))
 		goto found_field;
 	      else
 		{
@@ -872,7 +890,7 @@ seed_gobject_get_property (JSContextRef context,
 	    }
 	  g_base_info_unref ((GIBaseInfo *) info);
 	}
-      return 0;
+      return NULL;
     }
 found:
 
@@ -895,7 +913,8 @@ seed_gobject_set_property (JSContextRef context,
   GValue gval = { 0 };
   GType type;
   gchar *cproperty_name;
-  gint length;
+  gsize length;
+  gsize i, len;
 
   if (seed_next_gobject_wrapper || JSValueIsNull (context, value))
     return 0;
@@ -912,7 +931,6 @@ seed_gobject_set_property (JSContextRef context,
 
   if (!spec)
     {
-      gint i, len;
       len = strlen (cproperty_name);
       for (i = 0; i < len; i++)
 	{
@@ -923,7 +941,7 @@ seed_gobject_set_property (JSContextRef context,
 					   cproperty_name);
       if (!spec)
 	{
-	  return 0;
+	  return FALSE;
 	}
     }
 
@@ -935,7 +953,7 @@ seed_gobject_set_property (JSContextRef context,
   seed_gvalue_from_seed_value (context, value, type, &gval, exception);
   if (*exception)
     {
-      return 0;
+      return FALSE;
     }
 
   if (glib_message)
@@ -960,18 +978,18 @@ seed_gobject_set_property (JSContextRef context,
 static JSValueRef
 seed_gobject_constructor_convert_to_type (JSContextRef ctx,
 					  JSObjectRef object,
-					  JSType type,
-					  JSValueRef *exception)
+					  JSType type, JSValueRef * exception)
 {
   GType gtype;
+  gchar *as_string;
 
   if (type == kJSTypeString)
     {
       JSValueRef ret;
-      gchar *as_string;
       gtype = (GType) JSObjectGetPrivate (object);
 
-      as_string = g_strdup_printf("[gobject_constructor %s]", g_type_name (gtype));
+      as_string =
+	g_strdup_printf ("[gobject_constructor %s]", g_type_name (gtype));
       ret = seed_value_from_string (ctx, as_string, exception);
       g_free (as_string);
 
@@ -981,8 +999,10 @@ seed_gobject_constructor_convert_to_type (JSContextRef ctx,
 }
 
 JSStaticFunction gobject_static_funcs[] = {
-  {"__debug_ref_count", seed_gobject_ref_count, 0},
-  {"__property_type", seed_gobject_property_type, 0},
+  {"__debug_ref_count", seed_gobject_ref_count, 0}
+  ,
+  {"__property_type", seed_gobject_property_type, 0}
+  ,
   {0, 0, 0}
 };
 
@@ -1126,11 +1146,21 @@ JSClassDefinition struct_constructor_def = {
   NULL				/* Convert To Type */
 };
 
+/**
+ * seed_create_function:
+ * @ctx: A valid #SeedContext
+ * @name: The name of the function (used in exceptions).
+ * @func: A #SeedFunctionCallback to implement the function.
+ * @obj: The #SeedObject on which to put the function. 
+ *
+ * Creates a JavaScript object representing a first-class function; when
+ * the function is called from JavaScript, @func will be called. Places
+ * the created function as the property @name on @obj.
+ *
+ */
 void
 seed_create_function (JSContextRef ctx,
-		      gchar * name,
-		      gpointer func,
-		      JSObjectRef obj)
+		      gchar * name, gpointer func, JSObjectRef obj)
 {
   JSObjectRef oref;
 
@@ -1144,7 +1174,7 @@ seed_repl_expose (JSContextRef ctx, ...)
   va_list argp;
   void *expose;
   JSObjectRef arrayObj;
-  int i = 0;
+  guint i = 0;
   JSStringRef script;
   JSObjectRef seed = (JSObjectRef) seed_object_get_property (ctx,
 							     JSContextGetGlobalObject
@@ -1158,7 +1188,7 @@ seed_repl_expose (JSContextRef ctx, ...)
 
   while ((expose = va_arg (argp, void *)))
     {
-      g_print ("  Seed.debug_argv[%d] = %p\n", i, expose);
+      g_print ("  Seed.debug_argv[%u] = %p\n", i, expose);
       JSObjectSetPropertyAtIndex (ctx, arrayObj, i++, expose, NULL);
     }
 
@@ -1236,6 +1266,7 @@ seed_parse_args (int *argc, char ***argv)
 {
   GOptionContext *option_context;
   GOptionGroup *seed_group;
+
   GError *error = NULL;
   gboolean ret = TRUE;
 
@@ -1264,96 +1295,6 @@ seed_parse_args (int *argc, char ***argv)
   return ret;
 }
 
-/**
- * seed_init:
- * @argc: A reference to the number of arguments remaining to parse.
- * @argv: A reference to an array of string arguments remaining to parse.
- *
- * Initializes a new #SeedEngine. This involves initializing GLib, creating
- * an initial context with all of the default globals, and initializing 
- * various internal parts of Seed.
- *
- * This function should only be called once within a single Seed application.
- *
- * Return value: The newly created and initialized #SeedEngine.
- *
- */
-SeedEngine *
-seed_init (gint * argc, gchar *** argv)
-{
-  g_type_init ();
-  g_log_set_handler ("GLib-GObject", G_LOG_LEVEL_WARNING, seed_log_handler,
-		     0);
-
-  if ((argc != 0) && seed_parse_args (argc, argv) == FALSE)
-    {
-      SEED_NOTE (MISC, "failed to parse arguments.");
-      return false;
-    }
-
-  qname = g_quark_from_static_string ("js-type");
-  qprototype = g_quark_from_static_string ("js-prototype");
-  js_ref_quark = g_quark_from_static_string ("js-ref");
-
-  eng = (SeedEngine *) g_malloc (sizeof (SeedEngine));
-
-  context_group = JSContextGroupCreate ();
-
-  eng->context = JSGlobalContextCreateInGroup (context_group, NULL);
-  eng->global = JSContextGetGlobalObject (eng->context);
-  eng->group = context_group;
-  eng->search_path = NULL;
-
-  gobject_class = JSClassCreate (&gobject_def);
-  JSClassRetain (gobject_class);
-  gobject_method_class = JSClassCreate (&gobject_method_def);
-  JSClassRetain (gobject_method_class);
-  gobject_constructor_class = JSClassCreate (&gobject_constructor_def);
-  JSClassRetain (gobject_constructor_class);
-  gobject_named_constructor_class =
-    JSClassCreate (&gobject_named_constructor_def);
-  JSClassRetain (gobject_named_constructor_class);
-  gobject_signal_class = JSClassCreate (seed_get_signal_class ());
-  JSClassRetain (gobject_signal_class);
-  seed_callback_class = JSClassCreate (&seed_callback_def);
-  JSClassRetain (seed_callback_class);
-  seed_struct_constructor_class = JSClassCreate (&struct_constructor_def);
-  JSClassRetain (seed_struct_constructor_class);
-  gobject_init_method_class = JSClassCreate (&gobject_init_method_def);
-  JSClassRetain (gobject_init_method_class);
-
-  g_type_set_qdata (G_TYPE_OBJECT, qname, gobject_class);
-
-  seed_obj_ref = JSObjectMake (eng->context, NULL, NULL);
-  seed_object_set_property (eng->context, eng->global, "Seed", seed_obj_ref);
-  JSValueProtect (eng->context, seed_obj_ref);
-
-  g_irepository_require (g_irepository_get_default (), "GObject", NULL, 0, 0);
-  g_irepository_require (g_irepository_get_default (), "GIRepository",
-			 NULL, 0, 0);
-
-  seed_initialize_importer (eng->context, eng->global);
-
-  seed_init_builtins (eng, argc, argv);
-  seed_closures_init ();
-  seed_structs_init ();
-
-  seed_gtype_init (eng);
-
-
-
-  defaults_script =
-	  JSStringCreateWithUTF8CString ("Seed.include(\""SEED_PREFIX_PATH"extensions/Seed.js\");");
-
-  JSEvaluateScript (eng->context, defaults_script, NULL, NULL, 0, NULL);
-
-  base_info_info =
-    g_irepository_find_by_name (0, "GIRepository", "IBaseInfo");
-
-  return eng;
-}
-
-
 /**
  * seed_init_with_context_group:
  * @argc: A reference to the number of arguments remaining to parse.
@@ -1371,10 +1312,8 @@ seed_init (gint * argc, gchar *** argv)
  */
 SeedEngine *
 seed_init_with_context_group (gint * argc,
-			      gchar *** argv,
-			      JSContextGroupRef group)
+			      gchar *** argv, JSContextGroupRef group)
 {
-  // FIXME: this seems like unnecessary code duplication
 
   g_type_init ();
   g_log_set_handler ("GLib-GObject", G_LOG_LEVEL_WARNING, seed_log_handler,
@@ -1383,7 +1322,7 @@ seed_init_with_context_group (gint * argc,
   if ((argc != 0) && seed_parse_args (argc, argv) == FALSE)
     {
       SEED_NOTE (MISC, "failed to parse arguments.");
-      return false;
+      return FALSE;
     }
 
   qname = g_quark_from_static_string ("js-type");
@@ -1399,6 +1338,9 @@ seed_init_with_context_group (gint * argc,
   eng->group = context_group;
   eng->search_path = NULL;
 
+  function_proto = (JSObjectRef)
+    seed_simple_evaluate (eng->context, "Function.prototype", NULL);
+
   gobject_class = JSClassCreate (&gobject_def);
   JSClassRetain (gobject_class);
   gobject_method_class = JSClassCreate (&gobject_method_def);
@@ -1435,13 +1377,37 @@ seed_init_with_context_group (gint * argc,
 
   seed_gtype_init (eng);
 
+
   defaults_script =
-	  JSStringCreateWithUTF8CString ("Seed.include(\""SEED_PREFIX_PATH"extensions/Seed.js\");");
+    JSStringCreateWithUTF8CString ("Seed.include(\"" SEED_PREFIX_PATH
+				   "extensions/Seed.js\");");
+
   JSEvaluateScript (eng->context, defaults_script, NULL, NULL, 0, NULL);
-  JSStringRelease (defaults_script);
 
   base_info_info =
     g_irepository_find_by_name (0, "GIRepository", "IBaseInfo");
 
   return eng;
 }
+
+/**
+ * seed_init:
+ * @argc: A reference to the number of arguments remaining to parse.
+ * @argv: A reference to an array of string arguments remaining to parse.
+ *
+ * Initializes a new #SeedEngine. This involves initializing GLib, creating
+ * an initial context with all of the default globals, and initializing
+ * various internal parts of Seed.
+ *
+ * This function should only be called once within a single Seed application.
+ *
+ * Return value: The newly created and initialized #SeedEngine.
+ *
+ */
+SeedEngine *
+seed_init (gint * argc, gchar *** argv)
+{
+  context_group = JSContextGroupCreate ();
+
+  return seed_init_with_context_group (argc, argv, context_group);
+}
